# -*- coding: utf-8 -*-
"""This module contains methods for discovering Sonos devices on the
network."""

from __future__ import unicode_literals

import logging
import threading
import socket
import select
from textwrap import dedent
import time
import struct
import ifaddr

from . import config
from .utils import really_utf8

_LOG = logging.getLogger(__name__)

# pylint: disable=too-many-locals, too-many-branches


class StoppableThread(threading.Thread):
    """ Thread class with a stop() method. """

    def __init__(self, target, args):
        super().__init__(target=target, args=args)
        self._stop_event = threading.Event()

    def stop(self):
        """Ask the thread to stop."""
        self._stop_event.set()

    def stopped(self):
        """Returns True if stop() has been called."""
        return self._stop_event.is_set()


def _discover_thread(callback,
                     interval,
                     include_invisible,
                     interface_addr):
    """ Discover Sonos zones on the local network. """

    def create_socket(interface_addr=None):
        """ A helper function for creating a socket for discover purposes.

        Create and return a socket with appropriate options set for multicast.
        """

        _sock = socket.socket(
            socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        # UPnP v1.0 requires a TTL of 4
        _sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL,
                         struct.pack("B", 4))
        _sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        if interface_addr is not None:
            _sock.setsockopt(
                socket.IPPROTO_IP, socket.IP_MULTICAST_IF,
                socket.inet_aton(interface_addr))
        return _sock

    # pylint: disable=invalid-name
    PLAYER_SEARCH = dedent("""\
        M-SEARCH * HTTP/1.1
        HOST: 239.255.255.250:1900
        MAN: "ssdp:discover"
        MX: 1
        ST: urn:schemas-upnp-org:device:ZonePlayer:1
        """).encode('utf-8')
    BCAST_ADDR = "255.255.255.255"
    MCAST_GRP = "239.255.255.250"
    MCAST_PORT = 1900

    _sockets = {}
    # Use the specified interface, if any
    if interface_addr is not None:
        try:
            address = socket.inet_aton(interface_addr)
        except socket.error:
            raise ValueError("{0} is not a valid IP address string".format(
                interface_addr))
        _sockets[interface_addr] = create_socket(interface_addr)
        _LOG.debug("Sending discovery packets on default interface")
    else:
        # Find the local network addresses using ifaddr.
        addresses = [
            ip.ip
            for adapter in ifaddr.get_adapters()
            for ip in adapter.ips
            if ip.is_IPv4
            if ip.ip != "127.0.0.1"
        ]

        # Create a socket for each unique address found, and one for the
        # default multicast address
        for address in addresses:
            try:
                _sockets[address] = create_socket(address)
            except socket.error as e:
                _LOG.debug("Can't make a discovery socket for %s: %s: %s",
                           address, e.__class__.__name__, e)

    resend = time.monotonic()
    while not threading.current_thread().stopped():
        if resend < time.monotonic():
            for _addr, _sock in _sockets.items():
                try:
                    _LOG.debug("Sending discovery packets on %s", _addr)
                    _sock.sendto(
                        really_utf8(PLAYER_SEARCH), (MCAST_GRP, MCAST_PORT))
                    _sock.sendto(
                        really_utf8(PLAYER_SEARCH), (BCAST_ADDR, MCAST_PORT))
                except OSError:
                    _LOG.debug("Discovery failed on %s", _addr)

            resend = time.monotonic() + interval
            seen = set()

        wait_time = resend - time.monotonic()
        response, _, _ = select.select(
            list(_sockets.values()), [], [], max(0, wait_time))

        # Only Zone Players should respond, given the value of ST in the
        # PLAYER_SEARCH message. However, to prevent misbehaved devices
        # on the network disrupting the discovery process, we check that
        # the response contains the "Sonos" string; otherwise we keep
        # waiting for a correct response.
        #
        # Here is a sample response from a real Sonos device (actual numbers
        # have been redacted):
        # HTTP/1.1 200 OK
        # CACHE-CONTROL: max-age = 1800
        # EXT:
        # LOCATION: http://***.***.***.***:1400/xml/device_description.xml
        # SERVER: Linux UPnP/1.0 Sonos/26.1-76230 (ZPS3)
        # ST: urn:schemas-upnp-org:device:ZonePlayer:1
        # USN: uuid:RINCON_B8*************00::urn:schemas-upnp-org:device:
        #                                                     ZonePlayer:1
        # X-RINCON-BOOTSEQ: 3
        # X-RINCON-HOUSEHOLD: Sonos_7O********************R7eU

        for _sock in response:
            try:
                data, addr = _sock.recvfrom(1024)
                _LOG.debug(
                    'Received discovery response from %s: "%s"', addr, data
                )

                if b"Sonos" not in data:
                    continue

                # pylint: disable=not-callable
                zone = config.SOCO_CLASS(addr[0])
                if zone in seen:
                    continue

                seen.add(zone)

                if include_invisible or zone.is_visible:
                    callback(zone)

            # pylint: disable=broad-except
            except Exception as ex:
                _LOG.debug('Error handling discovery response, ex=%s', ex)

    for _sock in _sockets.values():
        _sock.close()


def discover_thread(callback,
                    interval=60,
                    include_invisible=False,
                    interface_addr=None,
                    *,
                    start=True):
    """ Return a started thread with a discovery callback. """
    thread = StoppableThread(
        target=_discover_thread,
        args=(callback, interval, include_invisible, interface_addr))
    if start:
        thread.start()
    return thread


def discover(timeout=5,
             include_invisible=False,
             interface_addr=None,
             all_households=False):
    """ Discover Sonos zones on the local network.

    Return a set of `SoCo` instances for each zone found.
    Include invisible zones (bridges and slave zones in stereo pairs if
    ``include_invisible`` is `True`. Will block for up to ``timeout`` seconds,
     after which return `None` if no zones found.

    Args:
        timeout (int, optional): block for this many seconds, at most.
            Defaults to 5.
        include_invisible (bool, optional): include invisible zones in the
            return set. Defaults to `False`.
        interface_addr (str or None): Discovery operates by sending UDP
            multicast datagrams. ``interface_addr`` is a string (dotted
            quad) representation of the network interface address to use as
            the source of the datagrams (i.e. it is a value for
            `socket.IP_MULTICAST_IF <socket>`). If `None` or not specified,
            all system interfaces will be tried. Defaults to `None`.
        all_households (bool, optional): wait for all replies to discover
            multiple households. If `False` or not specified, return only
            the first household found.
    Returns:
        set: a set of `SoCo` instances, one for each zone found, or else
            `None`.

    """

    found_zones = set()
    first_response = None
    start = time.monotonic()

    def callback(zone):
        nonlocal first_response

        if first_response is None:
            first_response = time.monotonic()

        if include_invisible:
            found_zones.update(zone.all_zones)
        else:
            found_zones.update(zone.visible_zones)

        if not all_households:
            thread.stop()

    thread = discover_thread(
        callback, 2, include_invisible, interface_addr, start=False)
    thread.start()
    while thread.is_alive() and not thread.stopped():
        if first_response is None:
            thread.join(timeout=1)
            if time.monotonic() > start + timeout:
                thread.stop()
        else:
            thread.join(timeout=first_response + 1 - time.monotonic())
            thread.stop()

    return found_zones or None


def any_soco():
    """Return any visible soco device, for when it doesn't matter which.

    Try to obtain an existing instance, or use `discover` if necessary.
    Note that this assumes that the existing instance has not left
    the network.

    Returns:
        SoCo: A `SoCo` instance (or subclass if `config.SOCO_CLASS` is set,
            or `None` if no instances are found
    """

    cls = config.SOCO_CLASS
    # pylint: disable=no-member, protected-access
    try:
        # Try to get the first pre-existing soco instance we know about,
        # as long as it is visible (i.e. not a bridge etc). Otherwise,
        # perform discovery (again, excluding invisibles) and return one of
        # those
        device = next(d for d in cls._instances[cls._class_group].values()
                      if d.is_visible)
    except (KeyError, StopIteration):
        devices = discover()
        return None if devices is None else devices.pop()

    return device


def by_name(name):
    """Return a device by name.

    Args:
        name (str): The name of the device to return.

    Returns:
        :class:`~.SoCo`: The first device encountered among all zone with the
            given player name. If none are found `None` is returned.
    """
    devices = discover(all_households=True)

    for device in (devices or []):
        if device.player_name == name:
            return device
    return None



# -*- coding: utf-8 -*-
# pylint: disable=not-context-manager,import-error,wrong-import-position

# NOTE: The pylint not-content-manager warning is disabled pending the fix of
# a bug in pylint. See https://github.com/PyCQA/pylint/issues/782

# Disable while we have Python 2.x compatability
# pylint: disable=useless-object-inheritance


"""Classes to handle Sonos UPnP Events and Subscriptions.

The `Subscription` class from this module will be used in
:py:mod:`soco.services` if config.EVENTS_MODULE is set
to point to this module.

Example:

    Run this code, and change your volume, tracks etc::

        from __future__ import print_function
        import logging
        logging.basicConfig()
        import soco
        from pprint import pprint

        from soco import events_twisted
        soco.config.EVENTS_MODULE = events_twisted
        from twisted.internet import reactor

        def print_event(event):
            try:
                pprint (event.variables)
            except Exception as e:
                pprint ('There was an error in print_event:', e)

        def main():
            # pick a device at random and use it to get
            # the group coordinator
            device = soco.discover().pop().group.coordinator
            print (device.player_name)
            sub = device.renderingControl.subscribe().subscription
            sub2 = device.avTransport.subscribe().subscription
            sub.callback = print_event
            sub2.callback = print_event

            def before_shutdown():
                sub.unsubscribe()
                sub2.unsubscribe()
                events_twisted.event_listener.stop()

            reactor.addSystemEventTrigger(
                'before', 'shutdown', before_shutdown)

        if __name__=='__main__':
            reactor.callWhenRunning(main)
            reactor.run()

.. _Deferred: https://twistedmatrix.com/documents/current/api/\
twisted.internet.defer.Deferred.html
.. _Failure: https://twistedmatrix.com/documents/current/api/\
twisted.python.failure.Failure.html

"""

from __future__ import unicode_literals

import sys
import logging

# Hack to make docs build without twisted installed
if "sphinx" in sys.modules:
    class Resource(object):  # pylint: disable=no-init
        """Fake Resource class to use when building docs"""
else:
    from twisted.internet import reactor
    from twisted.web.server import Site
    from twisted.web.resource import Resource
    import twisted.internet.error
    from twisted.internet import task, defer
    from twisted.python.failure import Failure
    from twisted.web.client import Agent, BrowserLikeRedirectAgent
    from twisted.web.http_headers import Headers

# Event is imported for compatibility with events.py
# pylint: disable=unused-import
from .events_base import Event  # noqa: F401

from .events_base import (  # noqa: E402
    EventNotifyHandlerBase, EventListenerBase, SubscriptionBase,
    SubscriptionsMap
)

from .exceptions import SoCoException  # noqa: E402

log = logging.getLogger(__name__)  # pylint: disable=C0103


class EventNotifyHandler(Resource, EventNotifyHandlerBase):
    """Handles HTTP ``NOTIFY`` Verbs sent to the listener server.
    Inherits from `soco.events_base.EventNotifyHandlerBase`.
    """
    isLeaf = True

    def __init__(self):
        super(EventNotifyHandler, self).__init__()
        # The SubscriptionsMapTwisted instance created when this module is
        # imported. This is referenced by
        # soco.events_base.EventNotifyHandlerBase.
        self.subscriptions_map = subscriptions_map

    def render_NOTIFY(self, request):  # pylint: disable=invalid-name
        """Serve a ``NOTIFY`` request by calling `handle_notification`
        with the headers and content.
        """
        headers = {}
        for header in request.requestHeaders.getAllRawHeaders():
            decoded_key = header[0].decode('utf8').lower()
            decoded_header = header[1][0].decode('utf8')
            headers[decoded_key] = decoded_header
        content = request.content.read()
        self.handle_notification(headers, content)
        return b'OK'

    # pylint: disable=no-self-use, missing-docstring
    def log_event(self, seq, service_id, timestamp):
        log.info(
            "Event %s received for %s service at %s", seq,
            service_id, timestamp)


class EventListener(EventListenerBase):
    """The Event Listener.

    Runs an http server which is an endpoint for ``NOTIFY``
    requests from Sonos devices. Inherits from
    `soco.events_base.EventListenerBase`.
    """

    def __init__(self):
        super(EventListener, self).__init__()
        #:  :py:class:`twisted.internet.tcp.Port`: set at `listen`
        self.port = None

    def listen(self, ip_address):
        """Start the event listener listening on the local machine at
        port 1400 (default). If this port is unavailable, the
        listener will attempt to listen on the next available port,
        within a range of 100.

        Make sure that your firewall allows connections to this port.

        This method is called by `soco.events_base.EventListenerBase.start`

        Handling of requests is delegated to an instance of the
        `EventNotifyHandler` class.

        Args:
            ip_address (str): The local network interface on which the server
                should start listening.
        Returns:
            int: The port on which the server is listening.

        Note:
            The port on which the event listener listens is configurable.
            See `config.EVENT_LISTENER_PORT`
        """
        factory = Site(EventNotifyHandler())
        for port_number in range(self.requested_port_number,
                                 self.requested_port_number + 100):
            try:
                if port_number > self.requested_port_number:
                    log.warning("Trying next port (%d)", port_number)
                # pylint: disable=no-member
                self.port = reactor.listenTCP(port_number, factory,
                                              interface=ip_address)
                break
            # pylint: disable=invalid-name
            except twisted.internet.error.CannotListenError as e:
                log.warning(e)
                continue

        if self.port:
            log.info("Event listener running on %s", (ip_address,
                                                      self.port.port))
            return self.port.port
        else:
            return None

    # pylint: disable=unused-argument
    def stop_listening(self, address):
        """Stop the listener."""
        port, self.port = self.port, None
        port.stopListening()


class Subscription(SubscriptionBase):
    """A class representing the subscription to a UPnP event.
    Inherits from `soco.events_base.SubscriptionBase`.
    """
    def __init__(self, service, event_queue=None):
        """
        Args:
            service (Service): The SoCo `Service` to which the subscription
                 should be made.
            event_queue (:class:`~queue.Queue`): A queue on which received
                events will be put. If not specified, a queue will be
                created and used.

        """
        super(Subscription, self).__init__(service, event_queue)
        #: :py:obj:`function`: callback function to be called whenever an
        #: `Event` is received. If it is set and is callable, the callback
        #: function will be called with the `Event` as the only parameter and
        #: the Subscription's event queue won't be used.
        self.callback = None
        # The SubscriptionsMapTwisted instance created when this module is
        # imported. This is referenced by soco.events_base.SubscriptionBase.
        self.subscriptions_map = subscriptions_map
        # The EventListener instance created when this module is imported.
        # This is referenced by soco.events_base.SubscriptionBase.
        self.event_listener = event_listener
        # Used to keep track of the auto_renew loop
        self._auto_renew_loop = None
        # Used to serialise calls to subscribe, renew and unsubscribe
        self._queue = []

    # pylint: disable=arguments-differ
    def subscribe(self, requested_timeout=None, auto_renew=False, strict=True):
        """Subscribe to the service.

        If requested_timeout is provided, a subscription valid for that number
        of seconds will be requested, but not guaranteed. Check
        `timeout` on return to find out what period of validity is
        actually allocated.

        This method calls `events_base.SubscriptionBase.subscribe`.

        Note:
            SoCo will try to unsubscribe any subscriptions which are still
            subscribed on program termination, but it is good practice for
            you to clean up by making sure that you call :meth:`unsubscribe`
            yourself.

        Args:
            requested_timeout(int, optional): The timeout to be requested.
            auto_renew (bool, optional): If `True`, renew the subscription
                automatically shortly before timeout. Default `False`.
            strict (bool, optional): If True and an Exception occurs during
                execution, the returned Deferred_ will fail with a Failure_
                which will be passed to the applicable errback (if any has
                been set by the calling code) or, if False, the Failure will
                be logged and the Subscription instance will be passed to
                the applicable callback (if any has
                been set by the calling code). Default `True`.

        Returns:
            Deferred_: A Deferred_ the result of which will be the
                Subscription instance and the subscription property of which
                will point to the Subscription instance.

        """
        subscribe = super(Subscription, self).subscribe
        return self._wrap(subscribe, strict, requested_timeout, auto_renew)

    def renew(self, requested_timeout=None, is_autorenew=False, strict=True):
        """renew(requested_timeout=None)
        Renew the event subscription.
        You should not try to renew a subscription which has been
        unsubscribed, or once it has expired.

        This method calls `events_base.SubscriptionBase.renew`.

        Args:
            requested_timeout (int, optional): The period for which a renewal
                request should be made. If None (the default), use the timeout
                requested on subscription.
            is_autorenew (bool, optional): Whether this is an autorenewal.
                Default 'False'.
            strict (bool, optional): If True and an Exception occurs during
                execution, the returned Deferred_ will fail with a Failure_
                which will be passed to the applicable errback (if any has
                been set by the calling code) or, if False, the Failure will
                be logged and the Subscription instance will be passed to
                the applicable callback (if any has
                been set by the calling code). Default `True`.

        Returns:
            Deferred_: A Deferred_ the result of which will be the
                Subscription instance and the subscription property of which
                will point to the Subscription instance.

        """
        renew = super(Subscription, self).renew
        return self._wrap(renew, strict, requested_timeout, is_autorenew)

    def unsubscribe(self, strict=True):
        """unsubscribe()
        Unsubscribe from the service's events.
        Once unsubscribed, a Subscription instance should not be reused

        This method calls `events_base.SubscriptionBase.unsubscribe`.

        Args:
            strict (bool, optional): If True and an Exception occurs during
                execution, the returned Deferred_ will fail with a Failure_
                which will be passed to the applicable errback (if any has
                been set by the calling code) or, if False, the Failure will
                be logged and the Subscription instance will be passed to
                the applicable callback (if any has
                been set by the calling code). Default `True`.

        Returns:
            Deferred_: A Deferred_ the result of which will be the
                Subscription instance and the subscription property of which
                will point to the Subscription instance.
        """
        unsubscribe = super(Subscription, self).unsubscribe
        return self._wrap(unsubscribe, strict)

    def _auto_renew_start(self, interval):
        """Starts the auto_renew loop."""
        self._auto_renew_loop = task.LoopingCall(self.renew, is_autorenew=True,
                                                 strict=False)
        # False means wait for the interval to elapse, rather than fire at once
        self._auto_renew_loop.start(interval, False)

    def _auto_renew_cancel(self):
        """Cancels the auto_renew loop"""
        if self._auto_renew_loop:
            self._auto_renew_loop.stop()
            self._auto_renew_loop = None

    # pylint: disable=no-self-use, too-many-branches, too-many-arguments
    def _request(self, method, url, headers, success):
        """Sends an HTTP request.

        Args:
            method (str): 'SUBSCRIBE' or 'UNSUBSCRIBE'.
            url (str): The full endpoint to which the request is being sent.
            headers (dict): A dict of headers, each key and each value being
                of type `str`.
            success (function): A function to be called if the
                request succeeds. The function will be called with a dict
                of response headers as its only parameter.

        """
        agent = BrowserLikeRedirectAgent(Agent(reactor))

        if headers:
            for k in headers.keys():
                header = headers[k]
                del headers[k]
                if isinstance(header, (list,)):
                    header = header[0]
                if not isinstance(header, (bytes, bytearray)):
                    header = header.encode('latin-1')
                    k = k.encode('latin-1')
                headers[k] = [header]

        args = (
            method.encode('latin-1'),
            url.encode('latin-1'),
            Headers(headers)
        )
        d = agent.request(*args)  # pylint: disable=invalid-name

        def on_success(response):  # pylint: disable=missing-docstring
            response_headers = {}
            for header in response.headers.getAllRawHeaders():
                decoded_key = header[0].decode('utf8').lower()
                decoded_header = header[1][0].decode('utf8')
                response_headers[decoded_key] = decoded_header
            success(response_headers)
            return self

        d.addCallback(on_success)
        return d

    def _wrap(self, method, strict, *args, **kwargs):

        """This is a wrapper for `Subscription.subscribe`, `Subscription.renew`
        and `Subscription.unsubscribe` which:

            * Returns a deferred, the result of which will be the`Subscription`
              instance.
            * Sets deferred.subscription to point to the `Subscription`
              instance so a calling function can access the Subscription
              instance immediately without registering a Callback and waiting
              for it to fire.
            * Converts an Exception into a twisted.python.failure.Failure.
            * If a Failure (including an Exception converted into a Failure)
              has occurred:

                * Cancels the Subscription (unless the Failure was caused by a
                  SoCoException upon subscribe).
                * On an autorenew, if the strict flag was set to False, calls
                  the optional self.auto_renew_fail method with the
                  Failure.
                * If the strict flag was set to True (the default), passes the
                  Failure to the next Errback for handling or, if the strict
                  flag was set to False, logs the Failure instead.

            * Calls the `subscribing` and `finished_subscribing` methods of
              self.subscriptions_map, so that `count` property of
              self.subscriptions_map includes pending subscriptions.
            * Serialises calls to the wrapped methods, so that, for example, a
              call to unsubscribe will not commence until a call to subscribe
              has completed.

        """
        action = method.__name__

        # pylint: disable=unused-argument
        def execute(result, method, *args, **kwargs):
            """Execute method
            """
            # Increment the counter of pending calls to Subscription.subscribe
            # if method is subscribe
            if method.__name__ == 'subscribe':
                self.subscriptions_map.subscribing()

            # Execute method
            return method(*args, **kwargs)

        def callnext():
            """Call the next deferred in the queue.
            """
            # If there is another deferred in the queue,
            # call it
            if self._queue:
                d = self._queue[0]  # pylint: disable=invalid-name
                d.callback(None)

        def handle_outcome(outcome):
            """A callback / errback to handle the outcome ofmethod,
            after it has been executed
            """
            # We start by assuming no Failure occurred
            failure = None

            if isinstance(outcome, Failure):
                failure = outcome
                # If a Failure or Exception occurred during execution of
                # subscribe, renew or unsubscribe, cancel it unless the
                # Failure or Exception was a SoCoException upon subscribe
                if failure.type != SoCoException or action == 'renew':
                    msg = "An Exception occurred. Subscription to" +\
                        " {}, sid: {} has been cancelled".format(
                            self.service.base_url +
                            self.service.event_subscription_url,
                            self.sid)
                    self._cancel_subscription(msg)
                # If we're not being strict, log the Failure
                if not strict:
                    msg = "Failure received in Subscription" +\
                        ".{} for Subscription to:\n{}, sid: {}: {}".format(
                            action,
                            self.service.base_url +
                            self.service.event_subscription_url,
                            self.sid,
                            str(failure))
                    log.exception(msg)
                    # If we're not being strict upon a renewal
                    # (e.g. an autorenewal) call the optional
                    # self.auto_renew_fail method, if it has been set
                    if action == 'renew':
                        if self.auto_renew_fail:
                            if hasattr(self.auto_renew_fail, '__call__'):
                                # pylint: disable=not-callable
                                self.auto_renew_fail(failure)

            # Decrement the counter of pending calls to Subscription.subscribe
            # if completed action was subscribe
            if action == 'subscribe':
                self.subscriptions_map.finished_subscribing()

            # Remove the previous deferred from the queue
            self._queue.pop(0)

            # And call the next deferred in the queue
            callnext()

            # If a Failure occurred and we're in strict mode, reraise it
            if failure and strict:
                failure.trap()

        # Create a deferred
        d = defer.Deferred()  # pylint: disable=invalid-name
        # Set its subscription property to refer to this Subscription
        d.subscription = self
        # Set the deferred to execute method, when the
        # deferred is called
        d.addCallback(execute, method, *args, **kwargs)
        # Add handle_outcome as both a callback and errback
        d.addBoth(handle_outcome)
        # Add the deferred to the queue
        self._queue.append(d)
        # If this is the only deferred in the queue,
        # call it
        if len(self._queue) == 1:
            callnext()
        # Return the deferred
        return d


class SubscriptionsMapTwisted(SubscriptionsMap):
    """Maintains a mapping of sids to `soco.events_twisted.Subscription`
    instances. Registers each subscription to be unsubscribed at exit.

    Inherits from `soco.events_base.SubscriptionsMap`.
    """
    def __init__(self):
        super(SubscriptionsMapTwisted, self).__init__()
        # A counter of calls to Subscription.subscribe
        # that have started but not completed. This is
        # to prevent the event listener from being stopped prematurely
        self._pending = 0

    def register(self, subscription):
        """Register a subscription by updating local mapping of sid to
        subscription and registering it to be unsubscribed at exit.

        Args:
            subscription(`soco.events_twisted.Subscription`): the subscription
                to be registered.

        """

        # Add the subscription to the local dict of subscriptions so it
        # can be looked up by sid
        self.subscriptions[subscription.sid] = subscription
        # Register subscription to be unsubscribed at exit if still alive
        # pylint: disable=no-member
        reactor.addSystemEventTrigger('before', 'shutdown',
                                      subscription.unsubscribe)

    def subscribing(self):
        """Called when the `Subscription.subscribe` method
        commences execution.
        """
        # Increment the counter
        self._pending += 1

    def finished_subscribing(self):
        """Called when the `Subscription.subscribe` method
        completes execution.
        """
        # Decrement the counter
        self._pending -= 1

    @property
    def count(self):
        """
        `int`: The number of active or pending subscriptions.
        """
        return len(self.subscriptions) + self._pending


subscriptions_map = SubscriptionsMapTwisted()  # pylint: disable=C0103
event_listener = EventListener()  # pylint: disable=C0103



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import mock
import pytest

from pysonos import SoCo
from pysonos.exceptions import SoCoUPnPException

IP_ADDR = '192.168.1.101'


@pytest.yield_fixture()
def moco():
    """A mock soco with fake services and hardcoded is_coordinator.

    Allows calls to services to be tracked. Should not cause any network
    access
    """
    services = (
        'AVTransport', 'RenderingControl', 'DeviceProperties',
        'ContentDirectory', 'ZoneGroupTopology')
    patchers = [mock.patch('pysonos.core.{}'.format(service))
                for service in services]
    for patch in patchers:
        patch.start()
    with mock.patch("pysonos.SoCo.is_coordinator",
                    new_callable=mock.PropertyMock) as is_coord:
        is_coord = True
        yield SoCo(IP_ADDR)
    for patch in reversed(patchers):
        patch.stop()


class TestMusicLibrary:

    def test_search_track_no_result(self, moco):
        moco.contentDirectory.reset_mock()
        # Browse returns an exception if the artist can't be found
        # <s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/"><s:Body><s:Fault><faultcode>s:Client</faultcode><faultstring>UPnPError</faultstring><detail><UPnPError xmlns="urn:schemas-upnp-org:control-1-0"><errorCode>701</errorCode></UPnPError></detail></s:Fault></s:Body></s:Envelope>
        moco.contentDirectory.Browse.side_effect = SoCoUPnPException(
            "No such object", "701", "error XML")

        result = moco.music_library.search_track("artist")

        assert len(result) == 0

        moco.contentDirectory.Browse.assert_called_once_with([
            ('ObjectID', 'A:ALBUMARTIST/artist/'),
            ('BrowseFlag', 'BrowseDirectChildren'),
            ('Filter', '*'),
            ('StartingIndex', 0),
            ('RequestedCount', 100000),
            ('SortCriteria', '')
        ])

    def test_search_track_no_artist_album_track(self, moco):
        moco.contentDirectory.reset_mock()
        # Browse returns an empty result set if artist and album and
        # track cannot be found
        moco.contentDirectory.Browse.side_effect = None
        moco.contentDirectory.Browse.return_value = {
            'NumberReturned': '0',
            'Result': '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/" xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"></DIDL-Lite>',
            'TotalMatches': '0',
            'UpdateID': '0'
        }

        result = moco.music_library.search_track("artist", "album", "track")

        assert len(result) == 0

        moco.contentDirectory.Browse.assert_called_once_with([
            ('ObjectID', 'A:ALBUMARTIST/artist/album:track'),
            ('BrowseFlag', 'BrowseDirectChildren'),
            ('Filter', '*'),
            ('StartingIndex', 0),
            ('RequestedCount', 100000),
            ('SortCriteria', '')
        ])

    def test_search_track_artist_albums(self, moco):
        moco.contentDirectory.reset_mock()
        moco.contentDirectory.Browse.side_effect = None
        moco.contentDirectory.Browse.return_value = {
            'NumberReturned': '2',
            'Result': '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:ns0="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/" xmlns:ns2="urn:schemas-upnp-org:metadata-1-0/upnp/"><container id="A:ALBUMARTIST/The%20Artist/First%20Album" parentID="A:ALBUMARTIST/The%20Artist" restricted="true"><dc:title>First Album</dc:title><ns2:class>object.container.album.musicAlbum</ns2:class><res protocolInfo="x-rincon-playlist:*:*:*">x-rincon-playlist:RINCON_000123456789001400#A:ALBUMARTIST/The%20Artist/First%20Album</res><dc:creator>The Artist</dc:creator><ns2:albumArtURI>/getaa?u=x-file-cifs%3a%2f%2fserver%2fThe%2520Artist%2fFirst%2520Album%2ftrack2.mp3&amp;v=432</ns2:albumArtURI></container><container id="A:ALBUMARTIST/The%20Artist/Second%20Album" parentID="A:ALBUMARTIST/The%20Artist" restricted="true"><dc:title>Second Album</dc:title><ns2:class>object.container.album.musicAlbum</ns2:class><res protocolInfo="x-rincon-playlist:*:*:*">x-rincon-playlist:RINCON_000123456789001400#A:ALBUMARTIST/The%20Artist/Second%20Album</res><dc:creator>The Artist</dc:creator><ns2:albumArtURI>/getaa?u=x-file-cifs%3a%2f%2fserver%2fThe%2520Artist%2fSecond%2520Album%2ftrack2.mp3&amp;v=432</ns2:albumArtURI></container></DIDL-Lite>',
            'TotalMatches': '2',
            'UpdateID': '0'
        }
        results = moco.music_library.search_track("The Artist")

        assert results.number_returned == 2

        album = results[0]
        assert album.title == 'First Album'
        assert album.item_id == 'A:ALBUMARTIST/The%20Artist/First%20Album'
        album = results[1]
        assert album.title == 'Second Album'
        assert album.item_id == 'A:ALBUMARTIST/The%20Artist/Second%20Album'

        moco.contentDirectory.Browse.assert_called_once_with([
            ('ObjectID', 'A:ALBUMARTIST/The%20Artist/'),
            ('BrowseFlag', 'BrowseDirectChildren'),
            ('Filter', '*'),
            ('StartingIndex', 0),
            ('RequestedCount', 100000),
            ('SortCriteria', '')
        ])

    def test_search_track_artist_album_tracks(self, moco):
        moco.contentDirectory.reset_mock()
        moco.contentDirectory.Browse.side_effect = None
        moco.contentDirectory.Browse.return_value = {
            'NumberReturned': '3',
            'Result': '<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" xmlns:ns0="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/" xmlns:ns1="urn:schemas-upnp-org:metadata-1-0/upnp/"><item id="S://server/The%20Artist/The%20Album/03%20-%20Track%20Title%201.mp3" parentID="A:ALBUMARTIST/The%20Artist/The%20Album" restricted="true"><res protocolInfo="x-file-cifs:*:audio/mpeg:*">x-file-cifs://server/The%20Artist/The%20Album/03%20-%20Track%20Title%201.mp3</res><ns1:albumArtURI>/getaa?u=x-file-cifs%3a%2f%2fserver%2fThe%2520Artist%2fThe%2520Album%2f03%2520-%2520Track%2520Title%25201.mp3&amp;v=432</ns1:albumArtURI><dc:title>Track Title 1</dc:title><ns1:class>object.item.audioItem.musicTrack</ns1:class><dc:creator>The Artist</dc:creator><ns1:album>The Album</ns1:album><ns1:originalTrackNumber>3</ns1:originalTrackNumber></item><item id="S://server/The%20Artist/The%20Album/04%20-%20Track%20Title%202.m4a" parentID="A:ALBUMARTIST/The%20Artist/The%20Album" restricted="true"><res protocolInfo="x-file-cifs:*:audio/mp4:*">x-file-cifs://server/The%20Artist/The%20Album/04%20-%20Track%20Title%202.m4a</res><ns1:albumArtURI>/getaa?u=x-file-cifs%3a%2f%2fserver%2fThe%2520Artist%2fThe%2520Album%2f04%2520-%2520Track%2520Title%25202.m4a&amp;v=432</ns1:albumArtURI><dc:title>Track Title 2</dc:title><ns1:class>object.item.audioItem.musicTrack</ns1:class><dc:creator>The Artist</dc:creator><ns1:album>The Album</ns1:album><ns1:originalTrackNumber>4</ns1:originalTrackNumber></item><item id="S://server/The%20Artist/The%20Album/05%20-%20Track%20Title%203.mp3" parentID="A:ALBUMARTIST/The%20Artist/The%20Album" restricted="true"><res protocolInfo="x-file-cifs:*:audio/mpeg:*">x-file-cifs://server/The%20Artist/The%20Album/05%20-%20Track%20Title%203.mp3</res><ns1:albumArtURI>/getaa?u=x-file-cifs%3a%2f%2fserver%2fThe%2520Artist%2fThe%2520Album%2f05%2520-%2520Track%2520Title%25203.mp3&amp;v=432</ns1:albumArtURI><dc:title>Track Title 3</dc:title><ns1:class>object.item.audioItem.musicTrack</ns1:class><dc:creator>The Artist</dc:creator><ns1:album>The Album</ns1:album><ns1:originalTrackNumber>5</ns1:originalTrackNumber></item></DIDL-Lite>',
            'TotalMatches': '3',
            'UpdateID': '0'
        }

        results = moco.music_library.search_track("The Artist", "The Album")

        assert len(results) == 3

        track = results[0]
        assert track.title == 'Track Title 1'
        assert track.item_id == 'S://server/The%20Artist/The%20Album/03%20-%20Track%20Title%201.mp3'
        track = results[1]
        assert track.title == 'Track Title 2'
        assert track.item_id == 'S://server/The%20Artist/The%20Album/04%20-%20Track%20Title%202.m4a'
        track = results[2]
        assert track.title == 'Track Title 3'
        assert track.item_id == 'S://server/The%20Artist/The%20Album/05%20-%20Track%20Title%203.mp3'

        moco.contentDirectory.Browse.assert_called_once_with([
            ('ObjectID', 'A:ALBUMARTIST/The%20Artist/The%20Album'),
            ('BrowseFlag', 'BrowseDirectChildren'),
            ('Filter', '*'),
            ('StartingIndex', 0),
            ('RequestedCount', 100000),
            ('SortCriteria', '')
        ])

    def test_soco_library_updating(self, moco):
        moco.contentDirectory.GetShareIndexInProgress.return_value = {
            'IsIndexing': '0'}
        assert not moco.music_library.library_updating
        moco.contentDirectory.reset_mock()
        moco.contentDirectory.GetShareIndexInProgress.return_value = {
            'IsIndexing': '1'}
        assert moco.music_library.library_updating

    def test_soco_start_library_update(self, moco):
        moco.contentDirectory.RefreshShareIndex.return_value = True
        assert moco.music_library.start_library_update()
        moco.contentDirectory.RefreshShareIndex.assert_called_with([
            ('AlbumArtistDisplayOption', ''),
        ])

    def test_soco_list_library_shares(self, moco):
        # Tests with 2, 1 and 0 library shares
        moco.contentDirectory.Browse.return_value = {
            'TotalMatches': '2',
            'NumberReturned': '2',
            'UpdateID': '0',
            'Result': ('<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" '
                       'xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" '
                       'xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/" '
                       'xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/">'
                       '<container id="S://share_host/music_01/Music/Lossless" '
                       'parentID="S:" restricted="true"><dc:title>//share_host/music_01/Music/Lossless'
                       '</dc:title><upnp:class>object.container</upnp:class><res '
                       'protocolInfo="x-rincon-playlist:*:*:*">x-rincon-playlist:RINCON_XXXXXXXXXXXXX1400'
                       '#S://share_host/music_01/Music/Lossless</res></container>'
                       '<container id="S://share_host_2/music_01" parentID="S:" restricted="true">'
                       '<dc:title>//share_host_2/music_01</dc:title><upnp:class>object.container'
                       '</upnp:class><res protocolInfo="x-rincon-playlist:*:*:*">x-rincon-playlist'
                       ':RINCON_XXXXXXXXXXXXX1400#S://share_host_2/music_01</res></container></DIDL-Lite>')
        }
        results = moco.music_library.list_library_shares()
        assert len(results) == 2
        assert '//share_host/music_01/Music/Lossless' in results
        assert '//share_host_2/music_01' in results

        moco.contentDirectory.Browse.return_value = {
            'TotalMatches': '1',
            'NumberReturned': '1',
            'UpdateID': '0',
            'Result': ('<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" '
                       'xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" '
                       'xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/" '
                       'xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/">'
                       '<container id="S://share_host/music_01/Music/Lossless" parentID="S:" restricted="true">'
                       '<dc:title>//share_host/music_01/Music/Lossless</dc:title>'
                       '<upnp:class>object.container</upnp:class><res protocolInfo="x-rincon-playlist:*:*:*">'
                       'x-rincon-playlist:RINCON_XXXXXXXXXXXXXX400#'
                       'S://share_host/music_01/Music/Lossless</res></container></DIDL-Lite>')
        }
        results = moco.music_library.list_library_shares()
        assert len(results) == 1
        assert '//share_host/music_01/Music/Lossless' in results

        moco.contentDirectory.Browse.return_value = {
            'TotalMatches': '0',
            'NumberReturned': '0',
            'UpdateID': '0',
            'Result': ('<DIDL-Lite xmlns:dc="http://purl.org/dc/elements/1.1/" '
                       'xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/" '
                       'xmlns:r="urn:schemas-rinconnetworks-com:metadata-1-0/" '
                       'xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"/>')
        }
        results = moco.music_library.list_library_shares()
        assert len(results) == 0

    def test_soco_delete_library_share(self, moco):
        share = '//host/share'
        moco.music_library.delete_library_share(share)
        moco.contentDirectory.DestroyObject.assert_called_once_with([
            ('ObjectID', 'S:' + share)
        ])



# -*- coding: utf-8 -*-
"""Module to test the data structure classes with pytest."""

from __future__ import unicode_literals

import pytest

from pysonos import data_structures
from pysonos.exceptions import DIDLMetadataError
from pysonos.xml import XML


def assert_xml_equal(left, right, explain=None):
    """Helper function for comparing XML elements.

    Causes useful information to be output under pytest as to the differences
    between elements

    Args
         left (Element): an Elementtree.Element to compare
         right (Element): an Element to compare it with

    Raises
        AssertionError: if the Elements do not match
    """
    def _build_explanation(left, right, explain):
        if left.tag != right.tag:
            explain.append('tag <%s> does not match tag <%s>' % (left.tag,
                                                                 right.tag))
        for name, value in left.attrib.items():
            if right.get(name) != value:
                explain.append(
                    '%s attribute of element <%s> does not match: %s=%r, %s=%r' %
                    (name, left.tag, name, value, name, right.get(name)))
        for name in right.attrib:
            if name not in left.attrib:
                explain.append(
                    'right element <%s> has attribute %s but left does not' %
                    (left.tag, name))
        if left.text != right.text:
            explain.append(
                'text for element <%s>: %r != %r' %
                (left.tag, left.text, right.text))
        if left.tail != right.tail:
            explain.append(
                'tail for element <%s>: %r != %r' %
                (left.tag, left.text, right.text))
        for i1, i2 in zip(left, right):
            _build_explanation(i1, i2, explain)
        return

    explain = []
    _build_explanation(left, right, explain)
    if explain != []:
        header = "Comparing XML elements %s and %s" % (left, right)
        assert False, header + '\n'.join(explain)


class TestResource():
    """Testing the Resource class."""

    def test_create_didl_resource_with_no_params(self):
        with pytest.raises(TypeError):
            res = data_structures.DidlResource()

    def test_create_didl_resource(self):
        res = data_structures.DidlResource('a%20uri', 'a:protocol:info:xx')
        assert res.uri == 'a%20uri'
        assert res.protocol_info == 'a:protocol:info:xx'

    def test_create_didl_resource_to_from_element(self, helpers):
        res = data_structures.DidlResource('a%20uri', 'a:protocol:info:xx',
                                           bitrate=3)
        elt = res.to_element()
        assert helpers.compare_xml(
            elt,
            XML.fromstring(
                b'<res bitrate="3" '
                b'protocolInfo="a:protocol:info:xx">a%20uri</res>'
            )
        )
        assert data_structures.DidlResource.from_element(elt) == res

    def test_didl_resource_to_dict(self):
        res = data_structures.DidlResource('a%20uri', 'a:protocol:info:xx')
        rez = res.to_dict()
        assert rez['uri'] == 'a%20uri'
        assert rez['protocol_info'] == 'a:protocol:info:xx'
        assert len(rez) == 12

    def test_didl_resource_to_dict_remove_nones(self):
        res = data_structures.DidlResource('a%20uri', 'a:protocol:info:xx')
        rez = res.to_dict(remove_nones=True)
        assert rez['uri'] == 'a%20uri'
        assert rez['protocol_info'] == 'a:protocol:info:xx'
        assert len(rez) == 2

    def test_didl_resource_from_dict(self):
        res = data_structures.DidlResource('a%20uri', 'a:protocol:info:xx')
        rez = data_structures.DidlResource.from_dict(res.to_dict())
        assert res == rez

    def test_didl_resource_from_dict_remove_nones(self):
        res = data_structures.DidlResource('a%20uri', 'a:protocol:info:xx')
        rez = data_structures.DidlResource.from_dict(
            res.to_dict(remove_nones=True))
        assert res == rez

    def test_didl_resource_eq(self):
        res = data_structures.DidlResource('a%20uri', 'a:protocol:info:xx')
        assert res != data_structures.DidlObject(
            title='a_title', parent_id='pid', item_id='iid')
        assert res is not None
        assert res == res


class TestDidlObject():
    """Testing the DidlObject base class."""

    didl_xml = """
    <item xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
      xmlns:dc="http://purl.org/dc/elements/1.1/"
      xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
      id="iid" parentID="pid" restricted="true">
        <dc:title>the_title</dc:title>
        <upnp:class>object</upnp:class>
        <dc:creator>a_creator</dc:creator>
        <desc id="cdudn"
          nameSpace="urn:schemas-rinconnetworks-com:metadata-1-0/">DUMMY</desc>
    </item>
    """

    def test_create_didl_object_with_no_params(self):
        with pytest.raises(TypeError):
            didl_object = data_structures.DidlObject()

    def test_create_didl_object_with_disallowed_params(self):
        with pytest.raises(ValueError) as excinfo:
            didl_object = data_structures.DidlObject(
                title='a_title', parent_id='pid', item_id='iid', bad_args='other')
        assert 'not allowed' in str(excinfo.value)

    def test_create_didl_object_with_good_params(self):
        didl_object = data_structures.DidlObject(
            title='a_title',
            parent_id='pid',
            item_id='iid',
            creator='a_creator',
            desc="dummy")
        assert didl_object is not None
        assert didl_object.title == 'a_title'
        assert didl_object.parent_id == 'pid'
        assert didl_object.item_id == 'iid'
        assert didl_object.creator == 'a_creator'
        assert didl_object.resources == []
        assert didl_object.desc == "dummy"

    def test_didl_object_from_wrong_element(self):
        # Using the wrong element
        elt = XML.fromstring("""<res>URI</res>""")
        with pytest.raises(DIDLMetadataError) as excinfo:
            didl_object = data_structures.DidlObject.from_element(elt)
        assert "Wrong element. Expected <item> or <container>, "
        "got <res> for class object" in str(excinfo.value)

    def test_didl_object_from_element(self):
        elt = XML.fromstring(self.didl_xml)
        didl_object = data_structures.DidlObject.from_element(elt)
        assert didl_object.title == 'the_title'
        assert didl_object.parent_id == 'pid'
        assert didl_object.item_id == 'iid'
        assert didl_object.creator == 'a_creator'
        assert didl_object.desc == 'DUMMY'
        assert didl_object.item_class == 'object'

    def test_didl_object_from_element_unoff_subelement(self):
        """Test that for a DidlObject created from an element with an
        unofficial .# specified sub class, that the sub class is
        completely ignored

        """
        elt = XML.fromstring(
            self.didl_xml.replace('object', 'object.#SubClass')
        )
        didl_object = data_structures.DidlObject.from_element(elt)
        assert didl_object.item_class == 'object'

    def test_didl_object_from_wrong_class(self):
        # mismatched upnp class
        bad_elt1 = XML.fromstring(
            """<item xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/"
              xmlns:dc="http://purl.org/dc/elements/1.1/"
              xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/"
              id="iid" parentID="pid" restricted="true">
                 <dc:title>the_title</dc:title>
                 <upnp:class>object.item</upnp:class>
                 <dc:creator>a_creator</dc:creator>
                 <desc id="cdudn"
                   nameSpace="urn:schemas-rinconnetworks-com:metadata-1-0/">
                   RINCON_AssociatedZPUDN
                 </desc>
               </item>
        """)
        with pytest.raises(DIDLMetadataError) as excinfo:
            didl_object = data_structures.DidlObject.from_element(bad_elt1)
        assert ("UPnP class is incorrect. Expected 'object', got 'object.item'"
                ) in str(excinfo.value)

    def test_didl_object_from_dict(self):
        didl_object = data_structures.DidlObject(
            title='a_title',
            parent_id='pid',
            item_id='iid',
            creator='a_creator',
            desc='dummy')
        the_dict = {
            'title': 'a_title',
            'parent_id': 'pid',
            'item_id': 'iid',
            'creator': 'a_creator',
            'restricted': True,
            'desc': 'dummy'
        }
        assert data_structures.DidlObject.from_dict(the_dict) == didl_object
        # adding in an attibute not in _translation should make no difference
        the_dict['creator'] = 'another_creator'
        assert data_structures.DidlObject.from_dict(the_dict) != didl_object
        # round trip
        assert data_structures.DidlObject.from_dict(the_dict).to_dict() == \
            the_dict

    def test_didl_object_from_dict_resources(self):
        resources_list = [data_structures.DidlResource('a%20uri',
                                                       'a:protocol:info:xx')]
        didl_object = data_structures.DidlObject(title='a_title',
                                                 parent_id='pid',
                                                 item_id='iid',
                                                 creator='a_creator',
                                                 desc='dummy',
                                                 resources=resources_list)
        the_dict = {
            'title': 'a_title',
            'parent_id': 'pid',
            'item_id': 'iid',
            'creator': 'a_creator',
            'restricted': True,
            'desc': 'dummy',
            'resources': [resource.to_dict() for resource in resources_list]
        }
        assert data_structures.DidlObject.from_dict(the_dict) == didl_object

    def test_didl_object_from_dict_resources_remove_nones(self):
        resources_list = [data_structures.DidlResource('a%20uri',
                                                       'a:protocol:info:xx')]
        didl_object = data_structures.DidlObject(title='a_title',
                                                 parent_id='pid',
                                                 item_id='iid',
                                                 creator='a_creator',
                                                 desc='dummy',
                                                 resources=resources_list)
        the_dict = {
            'title': 'a_title',
            'parent_id': 'pid',
            'item_id': 'iid',
            'creator': 'a_creator',
            'restricted': True,
            'desc': 'dummy',
            'resources': [resource.to_dict(remove_nones=True)
                          for resource in resources_list]
        }
        assert data_structures.DidlObject.from_dict(the_dict) == didl_object

    def test_didl_comparisons(self):
        didl_object_1 = data_structures.DidlObject(
            title='a_title', parent_id='pid', item_id='iid', creator='a_creator')
        didl_object_2 = data_structures.DidlObject(
            title='a_title', parent_id='pid', item_id='iid', creator='a_creator')
        # should be not the same, but equal!
        assert didl_object_1 is not didl_object_2
        assert didl_object_1 == didl_object_2
        didl_object_3 = data_structures.DidlObject(
            title='a_title',
            parent_id='pid',
            item_id='iid',
            creator='a__different_creator')
        assert didl_object_3 != didl_object_1

    def test_didl_object_to_dict(self):
        didl_object = data_structures.DidlObject(
            title='a_title', parent_id='pid', item_id='iid', creator='a_creator')
        the_dict = {
            'title': 'a_title',
            'parent_id': 'pid',
            'item_id': 'iid',
            'creator': 'a_creator',
            'restricted': True,
            'desc': 'RINCON_AssociatedZPUDN'
        }
        assert didl_object.to_dict() == the_dict
        # adding in an attibute not in _translation should make no difference
        didl_object.other = 'other'
        assert didl_object.to_dict() == the_dict
        # but changing on the other should
        didl_object.creator = 'another'
        assert didl_object.to_dict() != the_dict

    def test_didl_object_to_dict_resources(self):
        resources_list = [data_structures.DidlResource('a%20uri',
                                                       'a:protocol:info:xx')]
        didl_object = data_structures.DidlObject(title='a_title',
                                                 parent_id='pid',
                                                 item_id='iid',
                                                 creator='a_creator',
                                                 resources=resources_list)
        the_dict = {
            'title': 'a_title',
            'parent_id': 'pid',
            'item_id': 'iid',
            'creator': 'a_creator',
            'restricted': True,
            'desc': 'RINCON_AssociatedZPUDN',
            'resources': [resource.to_dict() for resource in resources_list]
        }
        assert didl_object.to_dict() == the_dict

    def test_didl_object_to_dict_resources_remove_nones(self):
        resources_list = [data_structures.DidlResource('a%20uri',
                                                       'a:protocol:info:xx')]
        didl_object = data_structures.DidlObject(title='a_title',
                                                 parent_id='pid',
                                                 item_id='iid',
                                                 creator='a_creator',
                                                 resources=resources_list)
        the_dict = {
            'title': 'a_title',
            'parent_id': 'pid',
            'item_id': 'iid',
            'creator': 'a_creator',
            'restricted': True,
            'desc': 'RINCON_AssociatedZPUDN',
            'resources': [resource.to_dict(remove_nones=True)
                          for resource in resources_list]
        }
        assert didl_object.to_dict(remove_nones=True) == the_dict

    def test_didl_object_to_element(self):
        didl_object = data_structures.DidlObject(
            title='a_title', parent_id='pid', item_id='iid', creator='a_creator')
        # we seem to have to go through this to get ElementTree to deal
        # with namespaces properly!
        elt = XML.fromstring(XML.tostring(didl_object.to_element(True)))
        elt2 = XML.fromstring(
            '<dummy xmlns="urn:schemas-upnp-org:metadata-1-0/DIDL-Lite/" ' +
            'xmlns:dc="http://purl.org/dc/elements/1.1/" ' +
            'xmlns:upnp="urn:schemas-upnp-org:metadata-1-0/upnp/">' +
            '<item id="iid" parentID="pid" restricted="true">' +
            '<dc:title>a_title</dc:title>' +
            '<dc:creator>a_creator</dc:creator>' +
            '<upnp:class>object</upnp:class><desc id="cdudn" ' +
            'nameSpace="urn:schemas-rinconnetworks-com:metadata-1-0/">' +
            'RINCON_AssociatedZPUDN</desc></item></dummy>')[0]
        assert_xml_equal(elt2, elt)



# -*- coding: utf-8 -*-
# pylint: disable=not-context-manager

# NOTE: The pylint not-content-manager warning is disabled pending the fix of
# a bug in pylint. See https://github.com/PyCQA/pylint/issues/782

# Disable while we have Python 2.x compatability
# pylint: disable=useless-object-inheritance


"""Classes to handle Sonos UPnP Events and Subscriptions.

The `Subscription` class from this module will be used in
:py:mod:`soco.services` unless config.EVENTS_MODULE is set
to point to :py:mod:`soco.events_twisted`, in which case
:py:mod:`soco.events_twisted.Subscription` will be used.
See the Example in :py:mod:`soco.events_twisted`.

Example:

    Run this code, and change your volume, tracks etc::

        from __future__ import print_function
        try:
            from queue import Empty
        except:  # Py2.7
            from Queue import Empty

        import logging
        logging.basicConfig()
        import soco
        from pprint import pprint
        from soco.events import event_listener
        # pick a device at random and use it to get
        # the group coordinator
        device = soco.discover().pop().group.coordinator
        print (device.player_name)
        sub = device.renderingControl.subscribe()
        sub2 = device.avTransport.subscribe()

        while True:
            try:
                event = sub.events.get(timeout=0.5)
                pprint (event.variables)
            except Empty:
                pass
            try:
                event = sub2.events.get(timeout=0.5)
                pprint (event.variables)
            except Empty:
                pass

            except KeyboardInterrupt:
                sub.unsubscribe()
                sub2.unsubscribe()
                event_listener.stop()
                break
"""

from __future__ import unicode_literals

import logging
import threading

import requests

from .compat import (
    BaseHTTPRequestHandler, URLError, socketserver, urlopen
)

# Event is imported so that 'from events import Events' still works
# pylint: disable=unused-import
from .events_base import Event  # noqa: F401

from .events_base import (
    EventNotifyHandlerBase, EventListenerBase, SubscriptionBase,
    SubscriptionsMap
)

from .exceptions import SoCoException

log = logging.getLogger(__name__)  # pylint: disable=C0103


class EventServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    """A TCP server which handles each new request in a new thread."""
    allow_reuse_address = True


class EventNotifyHandler(BaseHTTPRequestHandler, EventNotifyHandlerBase):
    """Handles HTTP ``NOTIFY`` Verbs sent to the listener server.
    Inherits from `soco.events_base.EventNotifyHandlerBase`.
    """
    def __init__(self, *args, **kwargs):
        # The SubscriptionsMap instance created when this module is imported.
        # This is referenced by soco.events_base.EventNotifyHandlerBase.
        self.subscriptions_map = subscriptions_map
        # super appears at the end of __init__, because
        # BaseHTTPRequestHandler.__init__ does not return.
        super(EventNotifyHandler, self).__init__(*args, **kwargs)

    def do_NOTIFY(self):  # pylint: disable=invalid-name
        """Serve a ``NOTIFY`` request by calling `handle_notification`
        with the headers and content.
        """
        headers = requests.structures.CaseInsensitiveDict(self.headers)
        content_length = int(headers['content-length'])
        content = self.rfile.read(content_length)
        self.handle_notification(headers, content)
        self.send_response(200)
        self.end_headers()

    # pylint: disable=no-self-use, missing-docstring
    def log_event(self, seq, service_id, timestamp):
        log.debug(
            "Event %s received for %s service on thread %s at %s", seq,
            service_id, threading.current_thread(), timestamp)

    def log_message(self, fmt, *args):  # pylint: disable=arguments-differ
        # Divert standard webserver logging to the debug log
        log.debug(fmt, *args)


class EventServerThread(threading.Thread):
    """The thread in which the event listener server will run."""

    def __init__(self, address):
        """
        Args:
            address (tuple): The (ip, port) address on which the server
                should listen.
        """
        super(EventServerThread, self).__init__()
        #: `threading.Event`: Used to signal that the server should stop.
        self.stop_flag = threading.Event()
        #: `tuple`: The (ip, port) address on which the server is
        #: configured to listen.
        self.address = address

    def run(self):
        """Start the server on `address`.
        Handling of requests is delegated to an instance of the
        `EventNotifyHandler` class.
        """
        server = EventServer(self.address, EventNotifyHandler)
        log.debug("Event listener running on %s", server.server_address)
        # Listen for events until told to stop
        while not self.stop_flag.is_set():
            server.handle_request()

    def stop(self):
        """Stop the server.
        """
        self.stop_flag.set()


class EventListener(EventListenerBase):
    """The Event Listener.

    Runs an http server in a thread which is an endpoint for ``NOTIFY``
    requests from Sonos devices. Inherits from
    `soco.events_base.EventListenerBase`.
    """

    def __init__(self):
        super(EventListener, self).__init__()
        #: `EventServerThread`: thread on which to run.
        self._listener_thread = None

    def listen(self, ip_address):
        """Start the event listener listening on the local machine at
        port 1400 (default).

        Make sure that your firewall allows connections to this port.

        This method is called by `soco.events_base.EventListenerBase.start`

        Args:
            ip_address (str): The local network interface on which the server
                should start listening.
        Returns:
            int: `requested_port_number`. Included for
                compatibility with `soco.events_twisted.EventListener.listen`

        Note:
            The port on which the event listener listens is configurable.
            See `config.EVENT_LISTENER_PORT`
        """
        address = (ip_address, self.requested_port_number)
        self._listener_thread = EventServerThread(address)
        self._listener_thread.daemon = True
        self._listener_thread.start()
        return self.requested_port_number

    def stop_listening(self, address):
        """Stop the listener."""
        # Signal the thread to stop before handling the next request
        self._listener_thread.stop()
        # Send a dummy request in case the http server is currently listening
        try:
            urlopen(
                'http://%s:%s/' % (address[0], address[1]))
        except URLError:
            # If the server is already shut down, we receive a socket error,
            # which we ignore.
            pass
        # wait for the thread to finish, with a timeout of one second
        # to ensure the main thread does not hang
        self._listener_thread.join(1)
        # check if join timed out and issue a warning if it did
        if self._listener_thread.isAlive():
            log.debug('Event Listener did not shutdown gracefully.')


class Subscription(SubscriptionBase):
    """A class representing the subscription to a UPnP event.
    Inherits from `soco.events_base.SubscriptionBase`.
    """
    def __init__(self, service, event_queue=None):
        """
        Args:
            service (Service): The SoCo `Service` to which the subscription
                 should be made.
            event_queue (:class:`~queue.Queue`): A queue on which received
                events will be put. If not specified, a queue will be
                created and used.
        """
        super(Subscription, self).__init__(service, event_queue)
        # Used to keep track of the auto_renew thread
        self._auto_renew_thread = None
        self._auto_renew_thread_flag = threading.Event()
        # The SubscriptionsMap instance created when this module is imported.
        # This is referenced by soco.events_base.SubscriptionBase.
        self.subscriptions_map = subscriptions_map
        # The EventListener instance created when this module is imported.
        # This is referenced by soco.events_base.SubscriptionBase.
        self.event_listener = event_listener
        # Used to stop race conditions, as autorenewal may occur from a thread
        self._lock = threading.Lock()

    # pylint: disable=arguments-differ
    def subscribe(self, requested_timeout=None, auto_renew=False, strict=True):
        """Subscribe to the service.

        If requested_timeout is provided, a subscription valid for that number
        of seconds will be requested, but not guaranteed. Check
        `timeout` on return to find out what period of validity is
        actually allocated.

        This method calls `events_base.SubscriptionBase.subscribe`.

        Note:
            SoCo will try to unsubscribe any subscriptions which are still
            subscribed on program termination, but it is good practice for
            you to clean up by making sure that you call :meth:`unsubscribe`
            yourself.

        Args:
            requested_timeout(int, optional): The timeout to be requested.
            auto_renew (bool, optional): If `True`, renew the subscription
                automatically shortly before timeout. Default `False`.
            strict (bool, optional): If True and an Exception occurs during
                execution, the Exception will be raised or, if False, the
                Exception will be logged and the Subscription instance will be
                returned. Default `True`.

        Returns:
            `Subscription`: The Subscription instance.

        """
        subscribe = super(Subscription, self).subscribe
        return self._wrap(subscribe, strict, requested_timeout, auto_renew)

    def renew(self, requested_timeout=None, is_autorenew=False, strict=True):
        """renew(requested_timeout=None)
        Renew the event subscription.
        You should not try to renew a subscription which has been
        unsubscribed, or once it has expired.

        This method calls `events_base.SubscriptionBase.renew`.

        Args:
            requested_timeout (int, optional): The period for which a renewal
                request should be made. If None (the default), use the timeout
                requested on subscription.
            is_autorenew (bool, optional): Whether this is an autorenewal.
                Default 'False'.
            strict (bool, optional): If True and an Exception occurs during
                execution, the Exception will be raised or, if False, the
                Exception will be logged and the Subscription instance will be
                returned. Default `True`.

        Returns:
            `Subscription`: The Subscription instance.

        """
        renew = super(Subscription, self).renew
        return self._wrap(renew, strict, requested_timeout, is_autorenew)

    def unsubscribe(self, strict=True):
        """unsubscribe()
        Unsubscribe from the service's events.
        Once unsubscribed, a Subscription instance should not be reused

        This method calls `events_base.SubscriptionBase.unsubscribe`.

        Args:
            strict (bool, optional): If True and an Exception occurs during
                execution, the Exception will be raised or, if False, the
                Exception will be logged and the Subscription instance will be
                returned. Default `True`.

        Returns:
            `Subscription`: The Subscription instance.

        """
        unsubscribe = super(Subscription, self).unsubscribe
        return self._wrap(unsubscribe, strict)

    def _auto_renew_start(self, interval):
        """Starts the auto_renew thread."""

        class AutoRenewThread(threading.Thread):
            """Used by the auto_renew code to renew a subscription from within
            a thread.
            """

            def __init__(self, interval, stop_flag, sub, *args, **kwargs):
                super(AutoRenewThread, self).__init__(*args, **kwargs)
                self.interval = interval
                self.subscription = sub
                self.stop_flag = stop_flag
                self.daemon = True

            def run(self):
                subscription = self.subscription
                stop_flag = self.stop_flag
                interval = self.interval
                while not stop_flag.wait(interval):
                    subscription.renew(is_autorenew=True, strict=False)

        auto_renew_thread = AutoRenewThread(
            interval, self._auto_renew_thread_flag, self)
        auto_renew_thread.start()

    def _auto_renew_cancel(self):
        """Cancels the auto_renew thread"""
        self._auto_renew_thread_flag.set()

    # pylint: disable=no-self-use, too-many-arguments
    def _request(self, method, url, headers, success):
        """Sends an HTTP request.

        Args:
            method (str): 'SUBSCRIBE' or 'UNSUBSCRIBE'.
            url (str): The full endpoint to which the request is being sent.
            headers (dict): A dict of headers, each key and each value being
                of type `str`.
            success (function): A function to be called if the
                request succeeds. The function will be called with a dict
                of response headers as its only parameter.

        """
        response = None
        try:
            response = requests.request(
                method, url, headers=headers, timeout=3)
        except requests.exceptions.RequestException:
            if method != "UNSUBSCRIBE":
                raise

        # Ignore "412 Client Error: Precondition Failed for url:"
        # from rebooted speakers.
        if response and response.status_code != 412:
            response.raise_for_status()

        if success:
            success(response.headers)

    def _wrap(self, method, strict, *args, **kwargs):

        """This is a wrapper for `Subscription.subscribe`, `Subscription.renew`
        and `Subscription.unsubscribe` which:

            * Returns the`Subscription` instance.
            * If an Exception has occurred:

                * Cancels the Subscription (unless the Exception was caused by
                  a SoCoException upon subscribe).
                * On an autorenew, if the strict flag was set to False, calls
                  the optional self.auto_renew_fail method with the
                  Exception. This method needs to be threadsafe.
                * If the strict flag was set to True (the default), reraises
                  the Exception or, if the strict flag was set to False, logs
                  the Exception instead.

            * Calls the wrapped methods with a threading.Lock, to prevent race
              conditions (e.g. to prevent unsubscribe and autorenew being
              called simultaneously).

        """

        action = method.__name__

        # A lock is used, because autorenewal occurs in
        # a thread
        with self._lock:

            try:
                method(*args, **kwargs)

            except Exception as exc:  # pylint: disable=broad-except
                # If an Exception occurred during execution of subscribe,
                # renew or unsubscribe, set the cancel flag to True unless
                # the Exception was a SoCoException upon subscribe
                cancel = action == 'renew' or not \
                    isinstance(exc, SoCoException)

                if cancel:
                    # If the cancel flag was set to true, cancel the
                    # subscription with an explanation.
                    msg = "An Exception occurred. Subscription to" +\
                        " {}, sid: {} has been cancelled".format(
                            self.service.base_url +
                            self.service.event_subscription_url,
                            self.sid)
                    self._cancel_subscription(msg)
                # If we're not being strict, log the Exception
                if not strict:
                    msg = "Exception received in Subscription." +\
                        "{} for Subscription to:\n{}, sid: {}".format(
                            action,
                            self.service.base_url +
                            self.service.event_subscription_url,
                            self.sid)
                    log.exception(msg)
                    # If we're not being strict upon a renewal
                    # (e.g. an autorenewal) call the optional
                    # self.auto_renew_fail method, if it has been set
                    if action == 'renew' and self.auto_renew_fail is not None:
                        if hasattr(self.auto_renew_fail, '__call__'):
                            # pylint: disable=not-callable
                            self.auto_renew_fail(exc)

                # If we're being strict, reraise the Exception
                else:
                    raise  # pylint: disable=raising-bad-type

            else:
                # Return the Subscription to the function that
                # called subscribe, renew or unsubscribe (unless an
                # Exception occurred and it was reraised above)
                return self  # pylint: disable=lost-exception


subscriptions_map = SubscriptionsMap()  # pylint: disable=C0103
event_listener = EventListener()  # pylint: disable=C0103
